@@ -21,9 +21,9 @@ let x = N.f A;;
 
 [%%expect{|
 type (_, _) t = A : ('a, 'a) t | B : string -> ('a, 'b) t
-Line _, characters 52-74:
-  ....................................................function
-     | B s -> s
+Lines 8-9, characters 52-13:
+8 | ....................................................function
+9 |    | B s -> s
 Warning 8: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 A
