@@ -62,8 +62,7 @@ let match_expect_extension (ext : Parsetree.extension) =
   match ext with
   | ({Asttypes.txt="expect"|"ocaml.expect"; loc = extid_loc}, payload) ->
     let invalid_payload () =
-      Location.raise_errorf ~loc:extid_loc
-        "invalid [%%%%expect payload]"
+      Location.raise_errorf ~loc:extid_loc "invalid [%%%%expect payload]"
     in
     let string_constant (e : Parsetree.expression) =
       match e.pexp_desc with
@@ -130,26 +129,9 @@ let split_chunks phrases =
   loop phrases [] []
 
 module Compiler_messages = struct
-  let print_loc ppf (loc : Location.t) =
-    let startchar = loc.loc_start.pos_cnum - loc.loc_start.pos_bol in
-    let endchar = loc.loc_end.pos_cnum - loc.loc_start.pos_bol in
-    Format.fprintf ppf "Line _";
-    if startchar >= 0 then
-      Format.fprintf ppf ", characters %d-%d" startchar endchar;
-    Format.fprintf ppf ":@.";
-    if startchar >= 0 then
-      begin match !Location.input_lexbuf with
-      | None -> ()
-      | Some lexbuf ->
-         Location.show_code_at_location ppf lexbuf loc
-      end;
-    ()
-
   let capture ppf ~f =
     Misc.protect_refs
-      [ R (Location.formatter_for_warnings , ppf)
-      ; R (Location.printer                , print_loc)
-      ]
+      [ R (Location.formatter_for_warnings, ppf) ]
       f
 end
 
@@ -250,6 +232,7 @@ let eval_expect_file _fname ~file_contents =
     let _ : bool =
       List.fold_left phrases ~init:true ~f:(fun acc phrase ->
         acc &&
+        let snap = Btype.snapshot () in
         try
           exec_phrase ppf phrase
         with exn ->
@@ -260,6 +243,7 @@ let eval_expect_file _fname ~file_contents =
               (Printexc.to_string exn)
               (Printexc.raw_backtrace_to_string bt)
           end;
+          Btype.backtrack snap;
           false
       )
     in
@@ -329,8 +313,11 @@ let process_expect_file fname =
   write_corrected ~file:corrected_fname ~file_contents correction
 
 let repo_root = ref None
+let keep_original_error_size = ref false
 
 let main fname =
+  if not !keep_original_error_size then
+    Clflags.error_size := 0;
   Toploop.override_sys_argv
     (Array.sub Sys.argv ~pos:!Arg.current
        ~len:(Array.length Sys.argv - !Arg.current));
@@ -347,7 +334,7 @@ let main fname =
         Clflags.no_std_include := true;
         Compenv.last_include_dirs := [Filename.concat dir "stdlib"]
   end;
-  Compmisc.init_path false;
+  Compmisc.init_path ();
   Toploop.initialize_toplevel_env ();
   Sys.interactive := false;
   process_expect_file fname;
@@ -357,10 +344,9 @@ module Options = Main_args.Make_bytetop_options (struct
   let set r () = r := true
   let clear r () = r := false
   open Clflags
-  let _absname = set Location.absname
-  let _I dir =
-    let dir = Misc.expand_directory Config.standard_library dir in
-    include_dirs := dir :: !include_dirs
+  let _absname = set absname
+  let _alert = Warnings.parse_alert_option
+  let _I dir = include_dirs := dir :: !include_dirs
   let _init s = init_file := Some s
   let _noinit = set noinit
   let _labels = clear classic
@@ -373,6 +359,7 @@ module Options = Main_args.Make_bytetop_options (struct
   let _noprompt = set noprompt
   let _nopromptcont = set nopromptcont
   let _nostdlib = set no_std_include
+  let _nopervasives = set nopervasives
   let _open s = open_modules := s :: !open_modules
   let _ppx _s = (* disabled *) ()
   let _principal = set principal
@@ -388,7 +375,7 @@ module Options = Main_args.Make_bytetop_options (struct
   let _no_strict_formats = clear strict_formats
   let _unboxed_types = set unboxed_types
   let _no_unboxed_types = clear unboxed_types
-  let _unsafe = set fast
+  let _unsafe = set unsafe
   let _unsafe_string = set unsafe_string
   let _version () = (* disabled *) ()
   let _vnum () = (* disabled *) ()
@@ -407,6 +394,9 @@ module Options = Main_args.Make_bytetop_options (struct
   let _dtimings () = profile_columns := [ `Time ]
   let _dprofile () = profile_columns := Profile.all_columns
   let _dinstr = set dump_instr
+  let _dcamlprimc = set keep_camlprimc_file
+  let _color = Misc.set_or_ignore color_reader.parse color
+  let _error_style = Misc.set_or_ignore error_style_reader.parse error_style
 
   let _args = Arg.read_arg
   let _args0 = Arg.read_arg0
@@ -419,6 +409,8 @@ let args =
     ( [ "-repo-root", Arg.String (fun s -> repo_root := Some s),
         "<dir> root of the OCaml repository. This causes the tool to use \
          the stdlib from the current source tree rather than the installed one."
+      ; "-keep-original-error-size", Arg.Set keep_original_error_size,
+        " truncate long error messages as the compiler would"
       ] @ Options.list
     )
 
@@ -427,7 +419,6 @@ let usage = "Usage: expect_test <options> [script-file [arguments]]\n\
 
 let () =
   Clflags.color := Some Misc.Color.Never;
-  Clflags.error_size := 0;
   try
     Arg.parse args main usage;
     Printf.eprintf "expect_test: no input file\n";
